/* PanoTools
 * Copyright (C) 2004  Douglas Wilkins <dwilkins42@users.sourceforge.net> (the "Author").
 * All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of the Author of the
 * Software shall not be used in advertising or otherwise to promote the
 * sale, use or other dealings in this Software without prior written
 * authorization from the Author.
 */

#include "config.h"

#include "dbgutils.h"

#include <string.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

#include "plugin-intl.h"

#include "filter.h"

/*  Constants  */

#define PROCEDURE_NAME   "PanoPlugin"

#define DATA_KEY_VALS    "panoplugin"
#define DATA_KEY_UI_VALS "panoplugin_ui"

#define PARASITE_KEY     "panoplugin-options"

/* These are defined for translation only */
/* Remember not to translate <image> */
#define TR_PANTOOLS N_("<Image>/Filters/Panorama Tools")
#define TR_PANCTRLS N_("<Image>/Filters/Pan Controls")

#define AUTHOR "Helmut Dersch <der@fh-furtwangen.de> Douglas Wilkins <dwilkins42@users.sourceforge.net>"

/*  Local function prototypes  */

static void query (void);
static void run (const gchar      *name,
		       gint        nparams,
		 const GimpParam  *param,
		       gint       *nreturn_vals, 
		       GimpParam **return_vals);

GimpPlugInInfo PLUG_IN_INFO = 
{
  NULL,				/* init_proc  */
  NULL,				/* quit_proc  */
  query,			/* query_proc */
  run,				/* run_proc   */
};

MAIN ()

static void query (void)
{
  DENTER;
  gchar *help_path;
  gchar *help_uri;

  static GimpParamDef params[] = 
  {
    {GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive"},
    {GIMP_PDB_IMAGE, "image", "Input image"},
    {GIMP_PDB_DRAWABLE, "drawable_id", "Input Drawable"}
  };

  static GimpParamDef return_vals[] = 
  {
    {GIMP_PDB_IMAGE, "image", "Returned Image"},
  };

  gimp_plugin_domain_register (PLUGIN_NAME, LOCALEDIR);

  help_path = g_build_filename (DATADIR, "help", NULL);
  help_uri = g_filename_to_uri (help_path, NULL, NULL);
  g_free (help_path);

  gimp_plugin_help_register ("http://panotools.sourceforge.net/panoplugin/help",
			     help_uri);

  gimp_install_procedure ("plug_in_correct",
			  "Correct Scanned Photographs",
			  "",
			  AUTHOR,
			  LONGVERSION,
			  VERSION,
			  N_("<Image>/Filters/Panorama Tools/Correct"),
			  "RGB",
			  GIMP_PLUGIN,
			  G_N_ELEMENTS (params),
			  G_N_ELEMENTS (return_vals), 
			  params, 
			  return_vals);
  gimp_install_procedure ("plug_in_perspect", 
			  "Software Shift Lens", 
			  "",
			  AUTHOR, 
			  LONGVERSION, 
			  VERSION, 
			  N_("<Image>/Filters/Panorama Tools/Perspective"),
			  "RGB", 
			  GIMP_PLUGIN, 
			  G_N_ELEMENTS (params),
			  G_N_ELEMENTS (return_vals), 
			  params, 
			  return_vals);
  gimp_install_procedure ("plug_in_remap", 
			  "Change Projection", 
			  "", 
			  AUTHOR,
			  LONGVERSION, 
			  VERSION, 
			  N_("<Image>/Filters/Panorama Tools/Remap"),
			  "RGB",
			  GIMP_PLUGIN, 
			  G_N_ELEMENTS (params),
			  G_N_ELEMENTS (return_vals), 
			  params, 
			  return_vals);
  gimp_install_procedure ("plug_in_adjust",
			  "Insert/Extract Image into/from Panorama", 
			  "",
			  AUTHOR, 
			  LONGVERSION, 
			  VERSION, 
			  N_("<Image>/Filters/Panorama Tools/Adjust"),
			  "RGB",
			  GIMP_PLUGIN, 
			  G_N_ELEMENTS (params),
			  G_N_ELEMENTS (return_vals), 
			  params, 
			  return_vals);
  gimp_install_procedure ("plug_in_right", 
			  "Pan", 
			  "", 
			  AUTHOR, 
			  LONGVERSION,
			  VERSION, 
			  N_("<Image>/Filters/Pan Controls/Move Right"),
			  "RGB", 
			  GIMP_PLUGIN,
			  G_N_ELEMENTS (params), 
			  G_N_ELEMENTS (return_vals),
			  params, 
			  return_vals);
  gimp_install_procedure ("plug_in_left",
			  "Pan",
			  "",
			  AUTHOR,
			  LONGVERSION,
			  VERSION,
			  N_("<Image>/Filters/Pan Controls/Move Left"),
			  "RGB",
			  GIMP_PLUGIN,
			  G_N_ELEMENTS (params),
			  G_N_ELEMENTS (return_vals), 
			  params, 
			  return_vals);
  gimp_install_procedure ("plug_in_up",
			  "Pan",
			  "",
			  AUTHOR,
			  LONGVERSION,
			  VERSION,
			  N_("<Image>/Filters/Pan Controls/Tilt Up"),
			  "RGB",
			  GIMP_PLUGIN,
			  G_N_ELEMENTS (params),
			  G_N_ELEMENTS (return_vals), 
			  params, 
			  return_vals);
  gimp_install_procedure ("plug_in_down",
			  "Pan",
			  "",
			  AUTHOR,
			  LONGVERSION,
			  VERSION,
			  N_("<Image>/Filters/Pan Controls/Tilt Down"),
			  "RGB",
			  GIMP_PLUGIN,
			  G_N_ELEMENTS (params),
			  G_N_ELEMENTS (return_vals), 
			  params, 
			  return_vals);
  gimp_install_procedure ("plug_in_zoomin",
			  "Pan",
			  "",
			  AUTHOR,
			  LONGVERSION,
			  VERSION,
			  N_("<Image>/Filters/Pan Controls/Zoom In"),
			  "RGB",
			  GIMP_PLUGIN,
			  G_N_ELEMENTS (params),
			  G_N_ELEMENTS (return_vals), 
			  params, 
			  return_vals);
  gimp_install_procedure ("plug_in_zoomout", 
			  "Pan", 
			  "", 
			  AUTHOR, 
			  LONGVERSION,
			  VERSION, 
			  N_("<Image>/Filters/Pan Controls/Zoom Out"),
			  "RGB", 
			  GIMP_PLUGIN,
			  G_N_ELEMENTS (params), 
			  G_N_ELEMENTS (return_vals),
			  params, 
			  return_vals);
  gimp_install_procedure ("plug_in_apply", 
			  "Pan", 
			  "", 
			  AUTHOR, 
			  LONGVERSION,
			  VERSION, 
			  N_("<Image>/Filters/Pan Controls/Apply Changes"),
			  "RGB", 
			  GIMP_PLUGIN,
			  G_N_ELEMENTS (params), 
			  G_N_ELEMENTS (return_vals),
			  params, 
			  return_vals);
  gimp_install_procedure ("plug_in_getpano",
			  "Pan",
			  "",
			  AUTHOR,
			  LONGVERSION,
			  VERSION,
			  N_("<Image>/Filters/Pan Controls/Get Panorama"),
			  "RGB",
			  GIMP_PLUGIN,
			  G_N_ELEMENTS (params),
			  G_N_ELEMENTS (return_vals), 
			  params, 
			  return_vals);
  gimp_install_procedure ("plug_in_increment", 
			  "Pan", 
			  "", 
			  AUTHOR, 
			  LONGVERSION,
			  VERSION, 
			  N_("<Image>/Filters/Pan Controls/Increment"),
			  "RGB", 
			  GIMP_PLUGIN,
			  G_N_ELEMENTS (params), 
			  G_N_ELEMENTS (return_vals),
			  params, 
			  return_vals);
  DEXIT;
}

static void run (const gchar      *name,
		gint        nparams,
		const GimpParam  *param, 
		gint       *nreturn_vals, 
		GimpParam **return_vals)
{
	DENTER;
	static GimpParam return_values[1];
	GimpPDBStatusType status = GIMP_PDB_EXECUTION_ERROR;
	TrformStr TrStr;
	Image src, dest;
	sPrefs spref;
	guint32 image_id; 
	GimpDrawable *gimp_drawable;
	GimpPixelRgn gimp_src, gimp_dest;
	GimpRunMode run_mode;
	int need_4bytes = FALSE;

	src.data = NULL;		// To check later
	dest.data = NULL;

	/*  Initialize i18n support  */
	bindtextdomain (GETTEXT_PACKAGE, LOCALEDIR);
#ifdef HAVE_BIND_TEXTDOMAIN_CODESET
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
#endif
	textdomain (GETTEXT_PACKAGE);

	run_mode = param[0].data.d_int32;
	image_id = param[1].data.d_image;  
	gimp_drawable = gimp_drawable_get (param[2].data.d_drawable);

	switch (run_mode)
	{
		case GIMP_RUN_INTERACTIVE:
			TrStr.mode = _interactive + _show_progress + _hostCanResize;
			break;
		case GIMP_RUN_NONINTERACTIVE:
		case GIMP_RUN_WITH_LAST_VALS:
			TrStr.mode = _useprefs + _show_progress + _hostCanResize;
			break;
		default:
			return;
			break;
	}

	TrStr.tool = _correct;
	DTRACE ("Plugin name is: %s", name);

	if (strcmp (name, "plug_in_correct") == 0)
	{
		TrStr.tool = _correct;
		need_4bytes = FALSE;
	}
	else if (strcmp (name, "plug_in_perspect") == 0)
	{
		TrStr.tool = _perspective;
		need_4bytes = FALSE;
	}
	else if (strcmp (name, "plug_in_remap") == 0)
	{
		TrStr.tool = _remap;
		need_4bytes = FALSE;
	}
	else if (strcmp (name, "plug_in_adjust") == 0)
	{
		TrStr.tool = _adjust;
		need_4bytes = TRUE;
	}
	else if (strcmp (name, "plug_in_right") == 0)
	{
		TrStr.tool = _panright;
		need_4bytes = TRUE;;
	}
	else if (strcmp (name, "plug_in_left") == 0)
	{
		TrStr.tool = _panleft;
		need_4bytes = TRUE;;
	}
	else if (strcmp (name, "plug_in_up") == 0)
	{
		TrStr.tool = _panup;
		need_4bytes = TRUE;;
	}
	else if (strcmp (name, "plug_in_down") == 0)
	{
		TrStr.tool = _pandown;
		need_4bytes = TRUE;;
	}
	else if (strcmp (name, "plug_in_zoomin") == 0)
	{
		TrStr.tool = _zoomin;
		need_4bytes = TRUE;;
	}
	else if (strcmp (name, "plug_in_zoomout") == 0)
	{
		TrStr.tool = _zoomout;
		need_4bytes = TRUE;;
	}
	else if (strcmp (name, "plug_in_apply") == 0)
	{
		TrStr.tool = _apply;
		need_4bytes = TRUE;;
	}
	else if (strcmp (name, "plug_in_getpano") == 0)
	{
		TrStr.tool = _getPano;
		need_4bytes = TRUE;;
	}
	else if (strcmp (name, "plug_in_increment") == 0)
	{
		TrStr.tool = _increment;
		need_4bytes = TRUE;;
	}

	src.width = gimp_drawable->width;
	src.height = gimp_drawable->height;
	src.bitsPerPixel = (need_4bytes ? 32 : 24);
	src.bytesPerLine = src.width * src.bitsPerPixel / 8;
	src.dataformat = _RGB;
	src.dataSize = src.height * src.bytesPerLine;
	src.data = (unsigned char **) mymalloc (src.dataSize);
	src.selection.top = 0;
	src.selection.left = 0;
	src.selection.right = src.width;
	src.selection.bottom = src.height;

	DTRACE ("Src: w %ld h %ld bpp %ld bpl %ld ds %ld", src.width,src.height ,src.bitsPerPixel ,src.bytesPerLine,src.dataSize); 

	if (src.data == NULL)
	{
		PrintError ("Not enough memory!");
		return;
	}

	gimp_pixel_rgn_init (&gimp_src, gimp_drawable, 0, 0,
			gimp_drawable->width,
			gimp_drawable->height, FALSE, FALSE);

	gimp_pixel_rgn_get_rect (&gimp_src, *src.data, 0, 0, src.width, src.height);

	if (need_4bytes)
		ThreeToFourBPP (&src);

	TrStr.src = &src;
	TrStr.dest = &dest;

	TrStr.interpolator = _poly3;
	TrStr.gamma = 1.0;

	filter_main (&TrStr, &spref);

	*nreturn_vals = 1;
	*return_vals = return_values;

	if (TrStr.success)
	{
		DTRACE ("Transform success");  
		DTRACE ("Dest: w %ld h %ld bpp %ld bpl %ld ds %ld", 
				TrStr.dest->width, 
				TrStr.dest->height , 
				TrStr.dest->bitsPerPixel ,
				TrStr.dest->bytesPerLine,
				TrStr.dest->dataSize); 
		gimp_image_undo_group_start(image_id);
		// if size unchanged: Exchange with src:
		if (TrStr.src->width == TrStr.dest->width
				&& TrStr.src->height == TrStr.dest->height)
		{
			// Get changeable dest region
			gimp_pixel_rgn_init (&gimp_dest, gimp_drawable, 0, 0,
					gimp_drawable->width,
					gimp_drawable->height, TRUE, TRUE);
			if (need_4bytes)
				FourToThreeBPP (&dest);
			gimp_pixel_rgn_set_rect (&gimp_dest, *dest.data, 0, 0, dest.width,
					dest.height);
		}
		else			// size differs
		{
			// What should we do now: check preferences
			if (spref.displayPart)	// Display cropped/framed image
			{
				// Get changeable dest region
				gimp_pixel_rgn_init (&gimp_dest, gimp_drawable, 0, 0,
						gimp_drawable->width,
						gimp_drawable->height, TRUE, TRUE);
				CopyImageData (&src, &dest);
				if (need_4bytes)
					FourToThreeBPP (&src);
				gimp_pixel_rgn_set_rect (&gimp_dest, *src.data, 0, 0, src.width,
						src.height);
			}			// end display cropped part
			if (spref.saveFile)
			{
				// Display full sized image
				// Change size of drawable
				gimp_image_resize (gimp_drawable_get_image
						(gimp_drawable->drawable_id), dest.width,
						dest.height, 0, 0);
				gimp_layer_resize (gimp_drawable->drawable_id, dest.width,
						dest.height, 0, 0);
				gimp_drawable_flush (gimp_drawable);
				gimp_drawable = gimp_drawable_get (gimp_drawable->drawable_id);

				gimp_pixel_rgn_init (&gimp_dest, gimp_drawable, 0, 0,
						dest.width, dest.height, TRUE, TRUE);
				if (need_4bytes)
					FourToThreeBPP (&dest);
				gimp_pixel_rgn_set_rect (&gimp_dest, *dest.data, 0, 0,
						dest.width, dest.height);
			}
		}			// size differs

		gimp_drawable_flush (gimp_drawable);
		gimp_drawable_merge_shadow (gimp_drawable->drawable_id, TRUE);

		gimp_drawable_update (gimp_drawable->drawable_id, 0, 0,
				gimp_drawable->width, gimp_drawable->height);

		gimp_drawable_detach (gimp_drawable);

		gimp_displays_flush ();
		gimp_image_undo_group_end(image_id);


		status = GIMP_PDB_SUCCESS;
	}				// success
	else
	{
		status = GIMP_PDB_EXECUTION_ERROR;
	}

	if (src.data != NULL)
		myfree ((void **) src.data);
	if (dest.data != NULL)
		myfree ((void **) dest.data);

	return_values[0].data.d_status = status;
	DEXIT;
}
